#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "srvipx.h"
#include "globals.h"
#include "dbtools.h"
#include "ttycmd.h"
#ifdef IPXBSD
#include <netipx/ipx.h>
#include "srvip.h"
#endif

#ifdef IPXNW
IPXECB *CreateIPXECB(int socket);
SPXECB *clnECB[NUM_CLN_ECB];
IPXECB *cnsECB[NUM_CNS_ECB];
IPXECB *mcstECB[NUM_MCST_ECB];

#define ListenForConnection(n) {\
  for (b=0; b<n; b++) if (!(clnECB[b]->ecb.inUseFlag)) {\
    clnECB[b]->hdr.dataStreamType=0;\
    clnECB[b]->ecb.fragmentCount = 2;\
    SPXListenForSequencedPacket(&clnECB[b]->ecb);\
  }\
}
#endif

void closeIPXClients (void);
int initIPXClient (int socket);

IPXAddress IncomingIPXfec;
extern int clientTimeout;

int IPXConsumerSocket = 0;        /* IPX connections */
int clnSocketNumber = 0;          /* SPX connections */
int IPXMCastConsumerSocket = 0;   /* IPX connections */

void closeIPXClients (void)
{
  int i;
  char str[256];
  if (!IPXloaded) return;

# ifdef IPXNW
  if (IPXConsumerSocket) IPXCloseSocket(IPXConsumerSocket);
  if (clnSocketNumber) IPXCloseSocket(clnSocketNumber);
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (FecTbl[conTbl[i]->fecIdx].TransportProtocol == IPX)
    {
      sprintf(str,"IPX link to %s %s terminated",
             conTbl[i]->expName, conTbl[i]->prpName);
      feclog(str);
    }
  }
# else
  if (IPXConsumerSocket)
  {
     close(IPXConsumerSocket);
     IPXConsumerSocket = 0;
  }
  if (ListenForMCasts && IPXMCastConsumerSocket)
  {
     close(IPXMCastConsumerSocket);
     IPXMCastConsumerSocket = 0;
  }
# endif
  return;
}

int initIPXClient(int socket)
{
  int i;

  if (!IPXloaded) return ipx_not_supported;
# ifdef IPXNW
  if (IPXOpenSocket((BYTE *)&socket,0)) return ipx_socket_error; /* open client socket */

  for (i=0; i<NUM_CNS_ECB; i++)           /* create consumer ECBs   */
  {
    if ((cnsECB[i]=CreateIPXECB(PCSocket)) == NULL) return out_of_client_memory;
    IPXListenForPacket(&cnsECB[i]->ecb);
  }

  if (nglobals)                 /* if FEC needs to listen for globals */
  {
    socket = *(int *)&brdcst_socket[0];
    if (IPXOpenSocket((BYTE *)&socket,0)) return ipx_socket_error;
    for (i=0; i<NUM_GLB_ECB; i++)
    {
      if ((glbECB[i]=CreateIPXECB(socket)) == NULL) return out_of_client_memory;
      IPXListenForPacket(&glbECB[i]->ecb);
    }
  }
  if (ListenForMCasts)           /* if FEC needs to listen for mcasts */
  {
    socket = *(int *)&mcast_socket[0];
    if (IPXOpenSocket((BYTE *)&socket,0)) return ipx_socket_error;
    for (i=0; i<NUM_MCST_ECB; i++)
    {
      if ((mcstECB[i]=CreateIPXECB(socket)) == NULL) return out_of_client_memory;
      IPXListenForPacket(&mcstECB[i]->ecb);
    }
  }
# else
  if (IPXConsumerSocket) close(IPXConsumerSocket);
  IPXConsumerSocket = ipxCreateSocket(0x0, socket);
  if (ListenForMCasts)
  {
    if (IPXMCastConsumerSocket) close(IPXMCastConsumerSocket);
    IPXMCastConsumerSocket = ipxCreateSocket(0x0, *(short *)mcast_socket);
    if (IPXMCastConsumerSocket <= 0) IPXMCastConsumerSocket = 0;
  }
# endif
  return 0;
}
void doIPXConsumer (void)
{
  int i;
  UINT32 prot;
# ifdef IPXBSD
  struct timeval ipxto;
  fd_set ipxconset;
  struct sockaddr_ipx IPXconsumer;
  int n, len = sizeof (IPXconsumer);
# endif

  if (!IPXloaded) return;
  IncomingNetworkProtocol = IPX;
  if (tineDebug) sprintf(fecprot, "IPX");

# ifdef IPXNW
  recvIncomingData(NUM_CNS_ECB,cnsECB,consumeData);
  if (nglobals) recvIncomingData(NUM_GLB_ECB,glbECB,getGlobalUpdate);
  if (ListenForMCasts) recvIncomingData(NUM_MCST_ECB,mcstECB,consumeData);
# else
  FD_ZERO(&ipxconset); FD_SET(IPXConsumerSocket, &ipxconset);
  if (IPXMCastConsumerSocket) FD_SET(IPXMCastConsumerSocket, &ipxconset);
  ipxto.tv_sec = 0; ipxto.tv_usec = 0;

  if (select(maxFdSets,&ipxconset,NULL,NULL,&ipxto) != -1)
  {
     recvIncomingIPXData(IPXConsumerSocket); n = 0;
     if (IPXMCastConsumerSocket) recvIncomingIPXData(IPXMCastConsumerSocket);
     FD_SET(IPXConsumerSocket, &ipxconset);
     if (IPXMCastConsumerSocket) FD_SET(IPXMCastConsumerSocket, &ipxconset);
  }
  else
  {
    if (errno != EWOULDBLOCK) if (tineDebug > 2)  soperror ("IPX select");
  }
  ReturnToClientCycle = 0;
# endif
  return;
}
