#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "srvip.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"

/**
 * \internal
 * \brief UDP server port.
 * 
 */
int gtUDPPort = UDP_PORT;

/**
 * \internal .
 * \brief UDP client port.
 */
int gtTCPPort = TCP_PORT;
                   
/**
 * \internal
 * \brief Address of primary ethernet card.
 */                                     
char gtEthrAddr[32]; /* INADDR_ANY if not given */

/**
  *\internal
 * \brief Port for multicast network services.
 */
int gtNetSrvPort = NETSRV_PORT;
          
/**
 * \internal
 * \brief Client UDP port for globals multicast.
 */
int gtGCastPort = GCAST_PORT;

/**
 * \internal
 * \brief Client UDP multicast port.
 */
int gtMCastPort = MCAST_PORT;

/**
 * \internal
 * \brief UDP client port.
 * 
 **/
int gtUDPClnPort = UDP_CLN_PORT;

/**
 * \internal
 * \brief Address of network services.
 */                                     
char *gtNetSrvAddr = NETSRV_ADDR;


/**
 * \internal
 * \brief Local broadcast address.
 */
char *gtNetLclBcast = NETBCST_ADDR;

/**
 * \internal
 * \brief Multicast address
*/
char *gtMCastAddr = MCAST_ADDR;
char gtMCastAddrBuffer[24];

/**
 * \internal
 * \brief Globals multicast address
 */
char *gtGCastAddr = GCAST_ADDR;
char *gtGCastTgtAddr = NULL;
char gtGCastAddrBuffer[24];
/**
 * \internal
 * \brief Time to live for multicasts (no. of router hops)
 */
int gtMCastTTL = MCAST_TTL;

#ifndef SMALL_TINE_LIB
TINE_EXPORT int SetSrvUDPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtUDPPort = port;
  return 0;
}
TINE_EXPORT int GetSrvUDPPort(void)
{
  return gtUDPPort;
}
TINE_EXPORT int SetSrvTCPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtTCPPort = port;
  return 0;
}
TINE_EXPORT int GetSrvTCPPort(void)
{
  return gtTCPPort;
}
TINE_EXPORT int SetGCastPort(int port)
{ 
  if (port < 1024) return not_allowed;
  gtGCastPort = port;
  return 0;
}
TINE_EXPORT int GetGCastPort(void)
{
  return gtGCastPort;
}
TINE_EXPORT int SetMCastPort(int port)
{
  if (port < 1024) return not_allowed;
  gtMCastPort = port;
  return 0;
}
TINE_EXPORT int GetMCastPort(void)
{
  return gtMCastPort;
}
TINE_EXPORT int SetClnUDPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtUDPClnPort = port;
  return 0;
}
TINE_EXPORT int GetClnUDPPort(void)
{
  return gtUDPClnPort;
}
TINE_EXPORT int SetNetSrvAddr(char *addr)
{
  static char mcaddr[32];
  char *c;
  int n;
  if (addr == NULL || strlen(addr) == 0) return argument_list_error;
  for (n=0; n<3; n++) 
  { /* must be of the form a.b.c.d */
    if ((c=strchr(addr,'.')) == NULL) break;
    c++;
  }
  if (n < 3) return not_allowed;
  strncpy(mcaddr,addr,24);
  gtNetSrvAddr = mcaddr;
  return 0;
}
TINE_EXPORT char *GetNetSrvAddr(void)
{
  return gtNetSrvAddr;
}
TINE_EXPORT int SetMCastAddr(char *addr)
{ /* sets the generic gcast address */
  char *c;
  int n;
  if (addr == NULL || strlen(addr) == 0) return argument_list_error;
  for (n=0; n<3; n++) 
  { /* must be of the form a.b.c.d */
    if ((c=strchr(addr,'.')) == NULL) break;
    c++;
  }
  if (n < 3) return not_allowed;
  strncpy(gtMCastAddrBuffer,addr,24);
  gtMCastAddr = gtMCastAddrBuffer;
  feclog("assign server multicast address to %s",addr);
  return 0;
}
TINE_EXPORT char *GetMCastAddr(void)
{
  return gtMCastAddr;
}
TINE_EXPORT int SetGCastAddr(char *addr)
{ /* sets the generic gcast address */
  char *c;
  int n;
  if (addr == NULL || strlen(addr) == 0) return argument_list_error;
  for (n=0; n<3; n++) 
  { /* must be of the form a.b.c.d */
    if ((c=strchr(addr,'.')) == NULL) break;
    c++;
  }
  if (n < 3) return not_allowed;
  strncpy(gtGCastAddrBuffer,addr,24);
  gtGCastAddr = gtGCastAddrBuffer;
  feclog("assign globals server multicast address to %s",addr);
  return 0;
}
TINE_EXPORT char *GetGCastAddr(void)
{
  return gtGCastAddr;
}
TINE_EXPORT int SetMCastTTL(int ttl)
{
  if (ttl < 1) return not_allowed;
  gtMCastTTL = ttl;
  return 0;
}
TINE_EXPORT int GetMCastTTL(void)
{
  return gtMCastTTL;
}
#endif /* not SMALL_TINE_LIB */

UINT32 TCPcount = 0, UDPcount = 0,
	     SPXcount = 0, IPXcount = 0,
	     BRDcount = 0;
char gClientAddrBuffer[30];
int TCPloaded = 0, IPXloaded = 0;
struct sockaddr_in prod_name;
fd_set ext_set;
#ifdef EXT_FDSET
int externalFdSets = -1;
#else
int externalFdSets = 0;
#endif
int maxFdSets = MAX_FDSETS;
void FD_MERGE(fd_set fdsrc,fd_set *fddst)
{
  int i;
  for (i=0; i<(int)sizeof(fd_set); i++)
    ((BYTE *)fddst)[i] |= ((BYTE *)&fdsrc)[i];
}
#ifdef WIN32
int isWindowsNT = TRUE;
int WinsockLoaded(void)
{
  static WSADATA wsaData;
  static int WSAinitialized = -1;
  static WORD wVersionRequested;
  char *ptr;

  if((ptr=getenv("OS")) != NULL && !stricmp(ptr,"Windows_NT"))
  {
    wVersionRequested = MAKEWORD( 2, 0 );
  }
  else
  {
    wVersionRequested = MAKEWORD( 1, 1 );
    isWindowsNT = FALSE;
  } 
 
	if (WSAinitialized == -1) WSAinitialized = WSAStartup(wVersionRequested,&wsaData) ? 0 : 1;
  return WSAinitialized;
}
int GetErrno(void) { return WSAGetLastError();}
#endif
#define MTU 1460

#ifndef MAX_RCVBUF
/**
 * \internal
 * \brief Default maximal number of receive buffers for socket options.
 */
# define MAX_RCVBUF 0x10000
#endif

#ifndef MAX_SNDBUF
/**
 * \internal
 * \brief Default maximal number of send buffers for socket options.
 *
 */
# define MAX_SNDBUF (MAX_RCVBUF / 2)
#endif

/**
 * \internal
 * \brief Maximal number of receive buffers for socket options.
 */
int max_rcvbuf = MAX_RCVBUF;

/**
 * \internal
 * \brief Maximal number of send buffers for socket options.
 */
int max_sndbuf = MAX_SNDBUF;

TINE_EXPORT int _SystemAssignBufferSpace(UINT32 rcvBufferSpace,UINT32 sndBufferSpace)
{
  if (rcvBufferSpace > 0)
  {
    if (rcvBufferSpace < 0x1000) return buffer_too_small; 
    max_rcvbuf = (int)rcvBufferSpace;
    feclog("socket recv buffer space set to %d bytes",rcvBufferSpace);
  }
  if (sndBufferSpace > 0)
  {
    if (sndBufferSpace < 0x1000) return buffer_too_small; 
    max_sndbuf = (int)sndBufferSpace;
    feclog("socket send buffer space set to %d bytes",sndBufferSpace);
  }
  return 0;
}

TCPBCKT *bcktList = NULL;

TCPBCKT *findBucket(int sck)
{
  TCPBCKT *bckt=NULL;
  for (bckt=bcktList; bckt != NULL && bckt->sck != sck; bckt=bckt->nxt);
  return bckt;
}
TCPBCKT *getBucket(int sck)
{
  TCPBCKT *bckt=NULL;
  for (bckt=bcktList; bckt != NULL; bckt=bckt->nxt)
  if (bckt->sck == sck) return bckt;
  if ((bckt=(TCPBCKT *)SystemCalloc(1,sizeof(TCPBCKT))) != NULL)
  {
    bckt->sck = sck;
    bckt->nxt = bcktList;
    bcktList = bckt;
  }
  return bckt;
}
void freeBucket(int sck)
{
  TCPBCKT *bckt,*b;
  if ((bckt=findBucket(sck)) == NULL) return;
  if (bckt == bcktList) bcktList = bckt->nxt;
  else
  {
    for (b=bcktList; b != NULL && b->nxt != bckt; b = b->nxt);
    if (b == NULL || b->nxt == NULL) return;
    b->nxt = bckt->nxt;
  }
  SystemFree(bckt);
}
#if defined(UNIX) && !defined(NO_TINE_PIPE)
void ClosePipeSocket(int sck)
{
  int i;
  unsigned int usck = (unsigned int)sck;

  close(sck);
  freeBucket(sck);

  for (i=0; i<nconsumers; i++) 
  { /* am I a server sending data to a client ? */
    if (Consumers[i]->sck == sck) Consumers[i]->sck = 0;
  }
  for (i=0; i<numFecTblEntries; i++)
  { /* am I a client trying to contact a server ? */
    if (!memcmp(&sck,FecTbl[i].IPXNetwork,4))
    {
      FecTbl[i].TransportProtocol = 0;
      memset(FecTbl[i].IPXNetwork,0,4);
      break;
    }
  }
}
#endif
void CloseConnectionSocket(int sck)
{
  int i;
  unsigned int usck = (unsigned int)sck;
  IDLE_CONNECTION *ic;

  freeBucket(sck);

  for (i=0; i<nTCPsck; i++)              
  { /* socket on the server side ? */
    if (sck == TCPsck[i])
    {
      if (sck != 0) soclose(sck);
      TCPsck[i] = 0;
      nTCPsck--;
      break;
    }
  }
  for (; i<nTCPsck; i++) TCPsck[i] = TCPsck[i+1];
  for (i=0; i<nconsumers; i++) if (Consumers[i]->sck == sck) Consumers[i]->sck = 0;
  for (i=0; i<nConnectionTableEntries; i++)
  { /* socket on the client side ? */
    if (conTbl[i]->TCPsocket == sck) /* socket on the client side ? */
    {
      closeIPConnectionToFec(conTbl[i]->fecIdx);
      while ((ic=getIdleConnectionBySocket(sck)) != NULL)
      { /* socket is now closed : it can't be idle ! */
        RemoveIdleConnection(ic);
      }
      return;
    }
  }
}
UINT32 sendtoErrcnt = 0;
int sendIPData(CLN *cln,BYTE *data)
{
  fd_set fdwset;
  struct timeval wait_to;
  int sck, cc = 0, retry = 0, nready;
  char *strprot = getProtocolAsString(cln->transport);
  unsigned short msgsize = PeekWord(*(unsigned short *)data);

  if (!TCPloaded) return tcp_not_supported;
  if (msgsize > UDP_BUFFER_SIZE) return use_stream_transport;
  if ((sck=cln->sck) <= 0) return tcp_socket_error;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (isUnixSocket(sck)) goto SendIPretry;
# endif
SelectIPretry:
  FD_ZERO(&fdwset); FD_SET(sck,&fdwset);
  wait_to.tv_sec=0;  wait_to.tv_usec=10000;
  if ((nready=select(maxFdSets,NULL,&fdwset,NULL,&wait_to)) != 1)
  {
    if (nready < 0) soperror("select");
    sendtoErrcnt++;
    if (retry++ > 3) ccerr(cannot_select);
    goto SelectIPretry;
  }
SendIPretry:
  if (writeto(sck,(char *)data,msgsize,0,
             (struct sockaddr *)&cln->IPaddress,sizeof(cln->IPaddress)) < 0)
  {
    sendtoErrcnt++;
    if (retry++ > 3) ccerr(net_write_error);
    goto SendIPretry;
  }
err:
  if (tineDebug)
  {
    inet_ntoa_b(cln->IPaddress.sin_addr,gClientAddrBuffer);
    if (cc)
    {
      soperror("sendto"); printf(">");
      dbglog("%s error: %s",strprot,erlst[cc]);
    }
    else if (tineDebug > 2)
    {
      dbglog("%s: %d bytes to %s [%s]",strprot,msgsize,cln->userName,gClientAddrBuffer);
    }
  }
  if (cc != 0)
  {
    if (cln->transport == TCP) 
    {
      CloseConnectionSocket(sck);
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    /* if (cln->transport == PIPE) ClosePipeSocket(sck); */ /* handled in sendSubRequest() */
#   endif
  }
  return cc;
}
struct in_addr tMCastAddr;
struct in_addr tMCastGenAddr;
struct in_addr tGCastAddr;
struct in_addr tGCastGenAddr;
struct in_addr tGCastTgtAddr;
struct in_addr tNetSrvAddr;
struct sockaddr_in tMCastAddrSck;
struct sockaddr_in tGCastAddrSck;
struct sockaddr_in tNetSrvAddrSck;
struct in_addr tEthrAddr; /* designated interface address (if multiple cards) */

int detachPortFromMulticastGroup(int sck,unsigned short port,struct in_addr grpAddr,char *grpType)
{
# ifndef NOT_MULTICASTING
  struct ip_mreq mreq;
  int i,n=0,cc;
  UINT32 haddr;

  memcpy(&haddr,&grpAddr,4);
  if (haddr == 0) 
  { /* not a valid group */
    return invalid_mcast_group;
  }
  if (sck == MCastConsumerSocket)
  {
    for (i=0; i<nConnectionTableEntries; i++)
    { /* count the links in the link table with this group */
      if (conTbl[i] == NULL) continue;
      if (conTbl[i]->mcastGrp == haddr) n++;
    }
  }
  else if (sck == GlobalSocket)
  {
    for (i=0; i<nglobals; i++) 
    { /* count the globals connected to this group */
      if (glbTbl[i]->mcastGrp == haddr) n++;
    }
  }
  if (n > 1)
  { /* there are other links attached to this group -> don't drop out ! */
    return 0;
  }
  mreq.imr_multiaddr.s_addr = grpAddr.s_addr;
  mreq.imr_interface.s_addr = strlen(gtEthrAddr) > 0 ? tEthrAddr.s_addr : htonl(INADDR_ANY);
  cc = setsockopt(sck,IPPROTO_IP,IP_DROP_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
  feclog("detaching from %s multicast group : %s",grpType,cc ? "failed" : "success");
  return cc ? mcast_init_error : 0;
# else
  sck = sck;
  port = port;
  grpAddr = grpAddr;
  grpType = grpType;
  return not_implemented;
# endif
}
int attachPortToMulticastGroup(int sck,unsigned short port,struct in_addr grpAddr,char *grpType)
{
# ifndef NOT_MULTICASTING
  struct ip_mreq mreq;
  int i,n=0,cc;
  UINT32 haddr;

  memcpy(&haddr,&grpAddr,4);
  if (haddr == 0) 
  { /* not a valid group */
    return invalid_mcast_group;
  }
  if (sck == MCastConsumerSocket)
  {
    for (i=0; i<nConnectionTableEntries; i++)
    { /* count the links in the link table with this group */
      if (conTbl[i] == NULL) continue;
      if (conTbl[i]->mcastGrp == haddr) n++;
    }
  }
  else if (sck == GlobalSocket)
  {
    for (i=0; i<nglobals; i++) 
    { /* count the globals connected to this group */
      if (glbTbl[i]->mcastGrp == haddr) n++;
    }
  }
  if (n > 0)
  { /* there are other links attached to this group -> no need to join ! */
    return 0;
  }
  if (grpType == NULL) grpType = "unknown";
  mreq.imr_multiaddr.s_addr = grpAddr.s_addr;
  mreq.imr_interface.s_addr = strlen(gtEthrAddr) > 0 ? tEthrAddr.s_addr : htonl(INADDR_ANY);
  cc = setsockopt(sck,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
  feclog("joining %s multicast group : %s",grpType,cc ? "failed" : "success");
  return cc ? mcast_init_error : 0;
# else
  sck = sck;
  port = port;
  grpAddr = grpAddr;
  grpType = grpType;
  return not_implemented;
# endif
}
#define MCAST_REATTACH_INTERVAL 300
int inhibitReattachToServiceGroup = TRUE;
void reAttachToServiceGroup(void)
{
  static time_t interim = 0;
  time_t t;
  if (inhibitReattachToServiceGroup) return;
  t = time(NULL);
  if (t < interim + MCAST_REATTACH_INTERVAL) return;
  attachPortToMulticastGroup(NetServiceSocket,(UINT16)gtNetSrvPort,tNetSrvAddr,"services");
  interim = t;
  return;
}
/**
 * \internal
 * \brief Binds an IP socket for UDP Communication and initializes some globals.
 *
 * Only binds a socket. Does <b>not</b> listen, as the name suggests.
 * If port equals one of the reserved port numbers gtMCastPort,
 * gtGCastPort, or gtNetSrvPort, the socket is added to a multicast group.
 * On UNIX and Win32 Systems, the send and receive buffers are set according
 * to the global variables max_sndbuf and max_rcvbuf.
 * 
 * Initializes the following globals when called for the first time:
 * <ul>
 * <li>tMCastAddr</li>
 * <li>tGCastAddr</li>
 * <li>tNetSrvAddr</li>
 * <li>tGCastAddrSck</li>
 * </ul>
 *
 * \param s Socket number (output)
 * \param port Port number. If set to 0, take default TINE client port.
 *
 * \todo separate initializer for globals
 * \todo rename (e.g. bindUdpSocket) (???)
 */
int listenUDP(int *s,unsigned short port,int initMultiCast)
{
  struct sockaddr_in sck_name;
  int mode = 1,optval,optlen, cc = 0;
  char txt[50];
  unsigned short udpport;
  static UINT16 nextFreePort = 0;
  static int lInitDone = FALSE;
  SINT32 ip, myip;
  struct in_addr inip;
# if defined(DOS) || defined(VXWORKS)
  int reuseFlg = TRUE;
# else
  int reuseFlg = FALSE;
# endif

  if (!lInitDone && initMultiCast) 
  { /* always fill this in at initialization time for use by server code */
    lInitDone = TRUE;
    myip = (SINT32)getmyipaddr();
#   if !defined(VXWORKS) && !defined(VMS) && !defined(NIOS) && !defined(DOS)
    ip = myip;
    ip |= 0xff000000;
#   else
    ip = (SINT32)(0xffffffff);
#   endif
#   ifdef NOT_MULTICASTING
    /* service port address */
    memcpy(&tNetSrvAddr,&ip,sizeof(struct in_addr));  
#   else
    /* canonical producer-subscriber address */
    ip = inet_addr(MCAST_ADDR);
    memcpy(&tMCastGenAddr,&ip,sizeof(struct in_addr));
    ip = inet_addr(gtMCastAddr);
    /* specific producer-subscriber address */
    if (gtMCastAddr != gtMCastAddrBuffer)
    { /* has not been pre-assigned */
      ((BYTE *)&ip)[3] = ((BYTE *)&myip)[3];
      ((BYTE *)&ip)[2] = ((BYTE *)&myip)[2];
      memcpy(&inip,&ip,4);
      inet_ntoa_b(inip,gtMCastAddrBuffer);
      gtMCastAddr = gtMCastAddrBuffer;
    }
    memcpy(&tMCastAddr,&ip,sizeof(struct in_addr));
    /* canonical producer-consumer address */
    ip = inet_addr(GCAST_ADDR);
    memcpy(&tGCastGenAddr,&ip,sizeof(struct in_addr));  
    ip = inet_addr(gtGCastAddr);
    /* specific producer-consumer address */
    if (gtGCastAddr != gtGCastAddrBuffer)
    { /* has not been pre-assigned */
      ((BYTE *)&ip)[3] = ((BYTE *)&myip)[3];
      ((BYTE *)&ip)[2] = ((BYTE *)&myip)[2];
      memcpy(&inip,&ip,4);
      inet_ntoa_b(inip,gtGCastAddrBuffer);
    }
    memcpy(&tGCastAddr,&ip,sizeof(struct in_addr));
    /* prepare return socket for gcasts from a server */
    tGCastAddrSck.sin_addr = tGCastAddr;
    tGCastAddrSck.sin_family = AF_INET;
    tGCastAddrSck.sin_port = htons((UINT16)gtGCastPort);
    /* service port address */
    ip = inet_addr(gtNetSrvAddr);
    memcpy(&tNetSrvAddr,&ip,sizeof(struct in_addr));
    if (strlen(gtEthrAddr) > 0)
    {
      ip = inet_addr(gtEthrAddr);
      memcpy(&tEthrAddr,&ip,sizeof(struct in_addr));  /* service port */
    }
#   endif
  }
  if (s == NULL) ccerr(argument_list_error);
  if (nextFreePort == 0) nextFreePort = gtUDPClnPort;
  sprintf(txt,"UDP PORT %d: ",port);
  if ((*s=(int)socket(PF_INET,SOCK_DGRAM,0)) < 0) soerr("socket");
  if (port == gtMCastPort || port == gtGCastPort || port == gtNetSrvPort) reuseFlg = TRUE;
  if (reuseFlg)
  {
    optval = 1; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) soerr("REUSEADDR");
  }
# if defined(UNIX) || defined(WIN32)
  if (max_rcvbuf > 0)
  {
    optval = max_rcvbuf; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_RCVBUF,(char *)&optval,optlen)) soerr("RCVBUF");
    if (getsockopt(*s,SOL_SOCKET,SO_RCVBUF,(char *)&optval,&optlen) == 0)
    {
      if (optval < max_rcvbuf) 
        feclog("requested %d bytes rcv buffer space, assigned %d bytes !",max_rcvbuf,optval);
    }
  }
  if (max_sndbuf)
  {
    optval = max_sndbuf; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_SNDBUF,(char *)&optval,optlen)) soerr("SNDBUF");
    if (getsockopt(*s,SOL_SOCKET,SO_SNDBUF,(char *)&optval,&optlen) == 0)
    {
      if (optval < max_sndbuf) 
        feclog("requested %d bytes snd buffer space, assigned %d bytes !",max_sndbuf,optval);
    }
  }
# endif
  memset(&sck_name,0,sizeof(sck_name));
  sck_name.sin_addr.s_addr = 0;
  sck_name.sin_family = AF_INET;

  udpport = port ? port : nextFreePort;
  if (udpport == 0) goto portbound;
bindport:
  sck_name.sin_port = htons(udpport);
  if (bind(*s,(struct sockaddr *)&sck_name,sizeof(sck_name)))
  {
    if (port == 0 && udpport++ < gtUDPClnPort + UDP_CLN_RANGE) goto bindport;
    soerr("bind");
  }
  if (port == 0) nextFreePort = udpport + 1;
portbound:
  /* attach to any relevant multicast groups (if possible) */
  if (port == gtNetSrvPort) attachPortToMulticastGroup(*s,port,tNetSrvAddr,"services");
  /* blocking or non-blocking ? */
  if (ServerCycleMode == POLLING && soioctl(*s,FIONBIO,(char *)&mode)) soerr("NONBLOCKING");
  feclog("%s bound to %d (socket %d %sblocking)",
          txt,ntohs(sck_name.sin_port),*s,ServerCycleMode == POLLING?" non-":" ");
err: 
  if (cc != 0)
  {
    soperror(txt); printf(">");
    if (s != NULL)
    {
      if (*s != 0) soclose(*s);
      *s = 0;
    }
    feclog("%s socket error",txt);
  }
  return cc;
}

#if !defined(VXWORKS) && !defined(VMS) && !defined(NIOS) && !defined(DOS)
CSVCOLUMN csvSrvIPColumns[] =
{
  { "ADDRESS","",0,CF_TEXT,16,(BYTE *)gtEthrAddr,NULL,0 },
};
CSVDB csvSrvIPDb =
{
  csvSrvIPColumns,
  sizeof(csvSrvIPColumns)/sizeof(csvSrvIPColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  0
};
int getSrvIPAddr(void)
{
  char *ptr;
  static int cc = 0;
  static int done = FALSE;

  if (done) return cc;
  gtEthrAddr[0] = 0;
  if ((cc=csvReadFile(FecDBpath,MY_ADDRESS_FILE,&csvSrvIPDb,NULL)) == 0) 
  {
    feclog("set fec address from file : %s",gtEthrAddr);
    done = TRUE;
    return cc;
  }
  if ((ptr=getenv("FEC_ADDRESS")) != NULL)
  {
    strncpy(gtEthrAddr,ptr,16);
    feclog("set fec address from environment : %s",gtEthrAddr);
    done = TRUE;
    return cc = 0;
  }
  cc = un_allocated;
  return cc;
}
UINT32 getmyipaddr(void)
{
  static char hname[30];
  static struct hostent *hent;
  static struct sockaddr_in stlclAddr;
  static struct sockaddr_in strmtAddr;
  static int done = 0;
  static UINT32 myip = 0;
  int sck;
  unsigned int nRet=sizeof(struct sockaddr);

  if (done) return myip;
# if defined(WIN32)
  /* do this here, just in case this is called prior to SystemInit() ! */
  if (!WinsockLoaded()) return -1;
# endif
  if (strlen(gtEthrAddr) > 0)
  {
    done = TRUE;
    return myip = inet_addr(gtEthrAddr);
  }
  stlclAddr.sin_addr.s_addr = INADDR_ANY;
  if (gethostname(hname,30) != -1)
  {
    if ((hent=gethostbyname(hname)) != NULL)
    {
      stlclAddr.sin_addr.s_addr = *((UINT32 *)(hent->h_addr));
    }
  }
  if (stlclAddr.sin_addr.s_addr == INADDR_ANY)
  {
    if ((sck=(int)socket(AF_INET, SOCK_DGRAM, 0)) > 0)
    {
      strmtAddr.sin_family = AF_INET;
      strmtAddr.sin_port = htons(IPPORT_ECHO);
      strmtAddr.sin_addr.s_addr = inet_addr("128.127.50.1");
      if (connect(sck,(struct sockaddr *)&strmtAddr,sizeof(struct sockaddr)) != -1)
      {
        getsockname(sck,(struct sockaddr *)&stlclAddr,(int *)&nRet);
      }
      soclose(sck);
    }
  }
  myip = stlclAddr.sin_addr.s_addr;
  done = TRUE;
  return stlclAddr.sin_addr.s_addr;  
}
#else
int getSrvIPAddr(void) { return not_implemented; }
#endif
/* trivial RPC call: */
int tUDPCall(int fecindex,BYTE *buf,int timout)
{
  fd_set tset;
  struct timeval t_to;
  static struct sockaddr_in IPfec;
  static CLN cln;
  static UINT16 ltid = 0xffff;
  UINT16 tid;
  int cc,n,poff;
  unsigned int len;
  char s[64];
  BYTE lclbuf[IPX_BUFFER_SIZE];
  FecDataStruct *fec = fecindex ? &FecTbl[fecindex] : &NameServer[curNameServer];

  if (SyncSocket == 0)
  {
    if (listenUDP(&SyncSocket,0,FALSE) != 0) 
    {
      soperror("Listen SyncSocket");
      return udp_socket_error;
    }
  }
  /* flush any input bytes that are somehow there ... */
  FD_ZERO(&tset); FD_SET(SyncSocket,&tset);
  t_to.tv_sec=0; t_to.tv_usec=0;
  while (select(32,&tset,NULL,NULL,&t_to) > 0)
  {
    len = sizeof(IPfec);
    recvfrom(SyncSocket,(char *)lclbuf,IPX_BUFFER_SIZE,0,(struct sockaddr *)&IPfec,(int *)&len);
  }
  memset(&cln,0,sizeof(CLN));
  strcpy(cln.userName,fec->Name);
  if (!fec->IPh_addr) return non_existent_fec;
  memset(&prod_name,0,sizeof(prod_name));
  memcpy(&prod_name.sin_addr,&fec->IPh_addr,sizeof(struct in_addr));
  poff = fec->PortOffset;
  prod_name.sin_family=AF_INET;
  prod_name.sin_port = htons((UINT16)(gtUDPPort + poff));
  memcpy(&cln.IPaddress,&prod_name,sizeof(prod_name));
  cln.sck = SyncSocket;

  if ((cc=sendIPData(&cln,buf)) != 0) return cc;

lsnForDataGram:
  FD_ZERO(&tset); FD_SET(SyncSocket,&tset);
  t_to.tv_sec=timout/1000;
# ifndef DOS
  t_to.tv_usec=((SINT32)timout%1000)*(SINT32)1000;
# else
  t_to.tv_usec=((SINT32)timout%1000);
# endif
  len = sizeof(IPfec);
  switch (select(maxFdSets,&tset,NULL,NULL,&t_to))
  {
    case -1: 
      if (tineDebug) 
      {
        sprintf(s,"SyncSocket (%d) select (to : %d)",SyncSocket,timout);
        soperror(s);
      }
      return tcp_socket_error;
    case 0:  
      return link_not_open;
    default: 
      n = recvfrom(SyncSocket,(char *)buf,IPX_BUFFER_SIZE,0,(struct sockaddr *)&IPfec,(int *)&len);
  }
  if (n != PeekWord(*(short *)buf)) return remitted_data_lost;
  if ((cc=PeekWord(*(short *)&buf[6])) == illegal_protocol) return cc;
  if ((tid=PeekWord(*(UINT16 *)&buf[4])) == ltid) /* phantom block ! */
  {
    goto lsnForDataGram;
  }
  ltid = tid;
  return 0;
}
#ifndef RDATE_TIMEOUT
# define RDATE_TIMEOUT 1
#endif
#define IPPORT_TIMESERVER	37              /* well known internet port number for rdate	*/
#define SEVENTY_YEARS		2208988800UL     /* midnight of 1.1.1900 to midnight of 1.1.1970	*/
char TimeServerHost[64];
#ifdef VXWORKS
struct hostent *gethostbyname(const char *hstname)
{
  static struct hostent hst;
  static char addr[4];
  int haddr;

  if ((haddr = hostGetByName((char *)hstname)) == ERROR) return NULL;

  hst.h_name = (char *)hstname;
  hst.h_aliases = 0;
  hst.h_addrtype = AF_INET;
  hst.h_length = strlen(hstname);
  memcpy(addr,&haddr,4);
  hst.h_addr = addr;

  return &hst;
}
struct hostent *gethostbyaddr(const char *hstaddr, int hstlen, int hsttype)
{
  static struct hostent hst;
  static char haddr[4];
  static char *addr; 
  char *c, hstaddrBuffer[32];
  int i = 0;

  strncpy(hstaddrBuffer,hstaddr,16); hstaddrBuffer[16] = 0;
  hst.h_name = (char *)hstaddr;
  hst.h_aliases = 0;
  hst.h_addrtype = hsttype;
  hst.h_length = hstlen;

  for (c=strtok(hstaddrBuffer,"."); c != NULL; c=strtok((char *)NULL,".")) 
  {
    haddr[i++] = atoi(c);
  }
  addr = (char *) &(haddr[0]);
  hst.h_addr = addr;

  return &hst;
}
#endif
time_t getUnixTime(char *tsHost)
{
  int sockfd=-1;																				/* socket fd's */
  static struct sockaddr_in  serverAddr;
  UINT32 tmptime;
  static struct hostent *host = NULL;
  static int firstCallLogged = 0;
  int cc = 0,mode = 1,done = 0;
  time_t interim;

  if (firstCallLogged == 0)
  {
    firstCallLogged++;
    feclog("using TIME Server %s", tsHost == NULL ? "local host" : tsHost);
  }
  if (tsHost == NULL) return time(NULL);
  if (host == NULL)
  {
    memset(&serverAddr,0,sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(IPPORT_TIMESERVER);
    if ((host=gethostbyname(tsHost)) == NULL &&
        (host=gethostbyaddr(tsHost,(int)strlen(tsHost),AF_INET)) == NULL)
    {
      if (tineDebug) soperror("time svr gethost");
      ccerr(non_existent);
    }
    memcpy(&serverAddr.sin_addr,host->h_addr,sizeof(struct in_addr));
  }

  if ((sockfd=(int)socket(AF_INET,SOCK_STREAM,0)) < 0 )
  {
    if (tineDebug) soperror("time svr socket");
    ccerr(tcp_socket_error);
  }
  if (soioctl(sockfd,FIONBIO,(char *)&mode))
  {
    if (tineDebug) soperror("time svr soioctl");
    ccerr(tcp_socket_error);
  }
  for (time(&interim);
       (connect(sockfd,(struct sockaddr *)&serverAddr,sizeof(serverAddr))<0);)
  {
    if (soerrno == EISCONN) break;
    if (time(NULL) > interim + RDATE_TIMEOUT)
    {
      if (tineDebug) soperror("time svr connect");
      ccerr(tcp_connect_error);
    }
  }

  while (time(NULL) < interim + RDATE_TIMEOUT && !done)
  {
    if (soread(sockfd,(char *)&tmptime,sizeof(UINT32)) == sizeof(UINT32) )
      done = 1;
  }
  if (!done)
  {
    if (tineDebug > 1) soperror("soread");
    ccerr(connection_timeout);
  }
  if (firstCallLogged == 1)
  {
    firstCallLogged++;
    feclog("got time from %s",tsHost);
  }

err:
  if (sockfd > 0) soclose(sockfd);
  if (cc)
  {
    if (firstCallLogged < 3)
    {
      firstCallLogged++;
      feclog("get time from %s : %s",tsHost,erlst[cc]);
    }
    return 0;
  }
  return (ntohl(tmptime) - SEVENTY_YEARS);  /* convert to unix time	*/
}
